# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['optga']

package_data = \
{'': ['*'], 'optga': ['tools/*']}

install_requires = \
['dataclasses_json>=0.3.7,<0.4.0',
 'numba>=0.46.0,<0.47.0',
 'numpy>=1.17,<2.0',
 'pandas>=0.25.3,<0.26.0']

entry_points = \
{'console_scripts': ['main = tests.test_optga:main']}

setup_kwargs = {
    'name': 'optga',
    'version': '0.1.1',
    'description': 'Framework to find out optimal input of machine learning model',
    'long_description': None,
    'author': 'horoiwa',
    'author_email': 'horoiwa195@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
